@@ -1,7 +1,7 @@
 package Perinci::CmdLine::Inline;
 
-our $DATE = '2017-01-11'; # DATE
-our $VERSION = '0.44'; # VERSION
+our $DATE = '2017-01-12'; # DATE
+our $VERSION = '0.45'; # VERSION
 
 use 5.010001;
 use strict;
@@ -68,6 +68,72 @@ sub _get_meta_from_url {
     }];
 }
 
+sub _gen_read_env {
+    my ($cd) = @_;
+    my @l2;
+
+    return "" unless $cd->{gen_args}{read_env};
+
+    _add_module($cd, "Complete::Bash");
+    push @l2, "{\n";
+    push @l2, '  last unless $_pci_r->{read_env};', "\n";
+    push @l2, '  my $env = $ENV{', dmp($cd->{gen_args}{env_name}), '};', "\n";
+    push @l2, '  last unless defined $env;', "\n";
+    push @l2, '  require Complete::Bash;', "\n";
+    push @l2, '  my ($words, undef) = @{ Complete::Bash::parse_cmdline($env, 0) };', "\n";
+    push @l2, '  unshift @ARGV, @$words;', "\n";
+    push @l2, "}\n";
+
+    join("", @l2);
+}
+
+sub _gen_read_config {
+    my ($cd) = @_;
+    my @l2;
+
+    return "" unless $cd->{gen_args}{read_config};
+
+    push @l2, 'if ($_pci_r->{read_config}) {', "\n";
+    _add_module($cd, "Perinci::CmdLine::Util::Config");
+    _add_module($cd, "Config::IOD::Reader"); # required by Perinci::CmdLine::Util::Config
+    _add_module($cd, "Config::IOD::Base"); # required by Config::IOD::Reader
+    _add_module($cd, "Data::Sah::Normalize"); # required by Perinci::CmdLine::Util::Config
+    _add_module($cd, "Perinci::Sub::Normalize"); # required by Perinci::CmdLine::Util::Config
+    _add_module($cd, "Sah::Schema::rinci::function_meta"); # required by Perinci::Sub::Normalize
+    push @l2, '  require Perinci::CmdLine::Util::Config;', "\n";
+    push @l2, "\n";
+    push @l2, '  my $res = Perinci::CmdLine::Util::Config::read_config(', "\n";
+    push @l2, '    config_paths     => $_pci_r->{config_paths},', "\n";
+    push @l2, '    config_filename  => ', dmp($cd->{gen_args}{config_filename}), ",\n";
+    push @l2, '    config_dirs      => ', dmp($cd->{gen_args}{config_dirs}), ' // ["$ENV{HOME}/.config", $ENV{HOME}, "/etc"],', "\n";
+    push @l2, '    program_name     => ', dmp($cd->{script_name}), ",\n";
+    push @l2, '  );', "\n";
+    push @l2, '  _pci_err($res) unless $res->[0] == 200;', "\n";
+    push @l2, '  $_pci_r->{config} = $res->[2];', "\n";
+    push @l2, '  $_pci_r->{read_config_files} = $res->[3]{"func.read_files"};', "\n";
+    push @l2, '  $_pci_r->{_config_section_read_order} = $res->[3]{"func.section_read_order"}; # we currently dont want to publish this request key', "\n";
+    push @l2, "\n";
+    push @l2, '  $res = Perinci::CmdLine::Util::Config::get_args_from_config(', "\n";
+    push @l2, '    r                  => $_pci_r,', "\n";
+    push @l2, '    config             => $_pci_r->{config},', "\n";
+    push @l2, '    args               => \%_pci_args,', "\n";
+    push @l2, '    program_name       => ', dmp($cd->{script_name}), ",\n";
+    push @l2, '    subcommand_name    => $_pci_r->{subcommand_name},', "\n";
+    push @l2, '    config_profile     => $_pci_r->{config_profile},', "\n";
+    push @l2, '    common_opts        => {},', "\n"; # XXX so currently we can't set e.g. format or
+    push @l2, '    meta               => $_pci_metas{ $_pci_r->{subcommand_name} },', "\n";
+    push @l2, '    meta_is_normalized => 1,', "\n";
+    push @l2, '  );', "\n";
+    push @l2, '  die $res unless $res->[0] == 200;', "\n";
+    push @l2, '  my $found = $res->[3]{"func.found"};', "\n";
+    push @l2, '  if (defined($_pci_r->{config_profile}) && !$found && defined($_pci_r->{read_config_files}) && @{$_pci_r->{read_config_files}} && !$_pci_r->{ignore_missing_config_profile_section}) {', "\n";
+    push @l2, '    _pci_err([412, "Profile \'$_pci_r->{config_profile}\' not found in configuration file"]);', "\n";
+    push @l2, '  }', "\n";
+    push @l2, '}', "\n"; # if read_config
+
+    join ("", @l2);
+}
+
 sub _gen_pci_check_args {
     my ($cd) = @_;
 
@@ -305,162 +371,225 @@ sub _gen_pci_check_args {
     $cd->{module_srcs}{"Local::_pci_check_args"} = "sub _pci_check_args {\n".join('', @l2)."}\n1;\n";
 }
 
-sub _gen_parse_opts {
+sub _gen_common_opt_handler {
+    my ($cd, $co) = @_;
+
+    my @l;
+
+    my $has_subcommands = $cd->{gen_args}{subcommands};
+
+    if ($co eq 'help') {
+        if ($has_subcommands) {
+            push @l, 'my $sc_name = $_pci_r->{subcommand_name}; ';
+            push @l, 'my $first_non_opt_arg; for (@ARGV) { next if /^-/; $first_non_opt_arg = $_; last } if (!length $sc_name && defined $first_non_opt_arg) { $sc_name = $first_non_opt_arg } ';
+            push @l, 'if (!length $sc_name) { print $help_msg } ';
+            for (sort keys %{ $cd->{helps} }) {
+                push @l, 'elsif ($sc_name eq '.dmp($_).') { print '.dmp($cd->{helps}{$_}).' } ';
+            }
+            push @l, 'else { _pci_err([500, "Unknown subcommand2: $sc_name"]) } ';
+            push @l, 'exit 0';
+        } else {
+            require Perinci::CmdLine::Help;
+            my $res = Perinci::CmdLine::Help::gen_help(
+                meta => $cd->{metas}{''},
+                common_opts => $cd->{copts},
+                program_name => $cd->{script_name},
+            );
+            return [500, "Can't generate help: $res->[0] - $res->[1]"]
+                unless $res->[0] == 200;
+            push @l, 'print ', dmp($res->[2]), '; exit 0;';
+        }
+    } elsif ($co eq 'version') {
+        no strict 'refs';
+        my $mod = $cd->{sc_mods}{''};
+        push @l, "no warnings 'once'; ";
+        push @l, "require $mod; " if $mod;
+        push @l, 'print "', $cd->{script_name} , ' version ", ',
+            (defined($cd->{gen_args}{script_version}) ? "\"$cd->{gen_args}{script_version}\"" :
+                 "(\$$mod\::VERSION // '?')"),
+                     ", (\$$mod\::DATE ? \" (\$$mod\::DATE)\" : '')",
+                         ', "\\n"; ';
+        push @l, 'print "  Generated by ', __PACKAGE__ , ' version ',
+            (${__PACKAGE__."::VERSION"} // 'dev'),
+                (${__PACKAGE__."::DATE"} ? " (".${__PACKAGE__."::DATE"}.")" : ""),
+                    '\n"; ';
+        push @l, 'exit 0';
+    } elsif ($co eq 'subcommands') {
+        my $scs_text = "Available subcommands:\n";
+        for (sort keys %{ $cd->{metas} }) {
+            $scs_text .= "  $_\n";
+        }
+        push @l, 'print ', dmp($scs_text), '; exit 0';
+    } elsif ($co eq 'cmd') {
+        push @l, '$_[2]{subcommand} = [$_[1]]; '; # for Getopt::Long::Subcommand
+        push @l, '$_pci_r->{subcommand_name} = $_[1];';
+    } elsif ($co eq 'format') {
+        push @l, '$_pci_r->{format} = $_[1];';
+    } elsif ($co eq 'json') {
+        push @l, '$_pci_r->{format} = (-t STDOUT) ? "json-pretty" : "json";';
+    } elsif ($co eq 'naked_res') {
+        push @l, '$_pci_r->{naked_res} = 1;';
+    } elsif ($co eq 'no_naked_res') {
+        push @l, '$_pci_r->{naked_res} = 0;';
+    } elsif ($co eq 'no_config') {
+        push @l, '        $_pci_r->{read_config} = 0;';
+    } elsif ($co eq 'config_path') {
+        push @l, '        $_pci_r->{config_paths} //= [];';
+        push @l, '        push @{ $_pci_r->{config_paths} }, $_[1];';
+    } elsif ($co eq 'config_profile') {
+        push @l, '        $_pci_r->{config_profile} = $_[1];';
+    } elsif ($co eq 'no_env') {
+        push @l, '        $_pci_r->{read_env} = 0;';
+    } else {
+        die "BUG: Unrecognized common_opt '$co'";
+    }
+    join "", @l;
+}
+
+sub _gen_get_args {
     my ($cd) = @_;
 
     my @l;
 
     push @l, 'my %mentioned_args;', "\n";
 
+    _add_module($cd, "Getopt::Long::EvenLess");
+    push @l, "require Getopt::Long::EvenLess;\n";
+
     if ($cd->{gen_args}{subcommands}) {
 
-        # generate help message for all subcommands
-        require Perinci::CmdLine::Help;
-        my %helps; # key = subcommand name
-        for my $sc_name (sort keys %{ $cd->{metas} }) {
-            next if $sc_name eq '';
-            my $meta = $cd->{metas}{$sc_name};
-            my $res = Perinci::CmdLine::Help::gen_help(
-                meta => $meta,
-                common_opts => { map {$_ => $cd->{copts}{$_}} grep { $_ !~ /\A(subcommands|cmd)\z/ } keys %{$cd->{copts}} },
-                program_name => "$cd->{script_name} $sc_name",
-            );
-            return [500, "Can't generate help (subcommand='$sc_name'): $res->[0] - $res->[1]"]
-                unless $res->[0] == 200;
-            $helps{$sc_name} = $res->[2];
-        }
-        # generate help when there is no subcommand specified
-        my $res = Perinci::CmdLine::Help::gen_help(
-            meta => $cd->{metas}{''},
-            common_opts => $cd->{copts},
-            program_name => $cd->{script_name},
-            subcommands => $cd->{gen_args}{subcommands},
-        );
-        return [500, "Can't generate help (subcommand=''): $res->[0] - $res->[1]"]
-            unless $res->[0] == 200;
-        $helps{''} = $res->[2];
-
         _add_module($cd, "Getopt::Long::Subcommand");
-        # we don't add the Complete::* that Getopt::Long::Subcommand depends to
-
-        push @l, 'my $help_msg = ', dmp($helps{''}), ";\n";
-
         push @l, "require Getopt::Long::Subcommand;\n";
-        push @l, 'my $res = Getopt::Long::Subcommand::GetOptions(', "\n";
+        # we haven't added the Complete::* that Getopt::Long::Subcommand depends on
+
+        # generate help message for all subcommands
+        {
+            require Perinci::CmdLine::Help;
+            my %helps; # key = subcommand name
+            for my $sc_name (sort keys %{ $cd->{metas} }) {
+                next if $sc_name eq '';
+                my $meta = $cd->{metas}{$sc_name};
+                my $res = Perinci::CmdLine::Help::gen_help(
+                    meta => $meta,
+                    common_opts => { map {$_ => $cd->{copts}{$_}} grep { $_ !~ /\A(subcommands|cmd)\z/ } keys %{$cd->{copts}} },
+                    program_name => "$cd->{script_name} $sc_name",
+                );
+                return [500, "Can't generate help (subcommand='$sc_name'): $res->[0] - $res->[1]"]
+                    unless $res->[0] == 200;
+                $helps{$sc_name} = $res->[2];
+            }
+            # generate help when there is no subcommand specified
+            my $res = Perinci::CmdLine::Help::gen_help(
+                meta => {v=>1.1},
+                common_opts => $cd->{copts},
+                program_name => $cd->{script_name},
+                program_summary => $cd->{gen_args}{script_summary},
+                subcommands => $cd->{gen_args}{subcommands},
+            );
+            return [500, "Can't generate help (subcommand=''): $res->[0] - $res->[1]"]
+                unless $res->[0] == 200;
+            $helps{''} = $res->[2];
+
+            $cd->{helps} = \%helps;
+        }
+
+        push @l, 'my $help_msg = ', dmp($cd->{helps}{''}), ";\n";
 
-        # common options
-        push @l, "  options => {\n";
         my @sc_names = sort keys %{ $cd->{metas} };
-        my $ggl_res = $cd->{ggl_res}{$sc_names[0]};
-        my $specmetas = $ggl_res->[3]{'func.specmeta'};
-        for my $o (sort keys %$specmetas) {
-            my $specmeta = $specmetas->{$o};
-            my $co = $specmeta->{common_opt};
-            next unless $co;
-            push @l, "    '$o' => {\n";
-            push @l, "      handler => sub {\n";
 
-            push @l, '        ';
-            if ($co eq 'help') {
-                push @l, 'my $sc_name = $_pci_r->{subcommand_name}; ';
-                push @l, 'if (!length $sc_name && @ARGV) { $sc_name = $ARGV[0] } ';
-                push @l, 'if (!length $sc_name && $_[2]{_non_options_argv} && @{ $_[2]{_non_options_argv} }) { $sc_name = $_[2]{_non_options_argv}[0] } ';
-                push @l, 'if (!length $sc_name) { print $help_msg } ';
-                for (sort keys %helps) {
-                    push @l, 'elsif ($sc_name eq '.dmp($_).') { print '.dmp($helps{$_}).' } ';
-                }
-                push @l, 'else { _pci_err([500, "Unknown subcommand2: $sc_name"]) } ';
-                push @l, 'exit 0', "\n";
-            } elsif ($co eq 'version') {
-                # begin B, duplicates A
-                no strict 'refs';
-                my $mod = $cd->{sc_mods}{''};
-                push @l, "        no warnings 'once';\n";
-                push @l, "        require $mod;\n" if $mod;
-                push @l, '        print "', $cd->{script_name} , ' version ", ',
-                    (defined($cd->{gen_args}{script_version}) ? "\"$cd->{gen_args}{script_version}\"" :
-                     "(\$$mod\::VERSION // '?')"),
-                     ", (\$$mod\::DATE ? \" (\$$mod\::DATE)\" : '')",
-                     ', "\\n";', "\n";
-                push @l, '        print "  Generated by ', __PACKAGE__ , ' version ',
-                    (${__PACKAGE__."::VERSION"} // 'dev'),
-                    (${__PACKAGE__."::DATE"} ? " (".${__PACKAGE__."::DATE"}.")" : ""),
-                    '\n";', "\n";
-                push @l, '        exit 0;', "\n";
-                # end B
-            } elsif ($co eq 'subcommands') {
-                my $scs_text = "Available subcommands:\n";
-                for (sort keys %{ $cd->{metas} }) {
-                    $scs_text .= "  $_\n";
-                }
-                push @l, 'print ', dmp($scs_text), "; exit 0;\n";
-            } elsif ($co eq 'cmd') {
-                push @l, '$_[2]{subcommand} = [$_[1]]; '; # for Getopt::Long::Subcommand
-                push @l, '$_pci_r->{subcommand_name} = $_[1];', "\n";
-            } elsif ($co eq 'format') {
-                push @l, '$_pci_r->{format} = $_[1];', "\n";
-            } elsif ($co eq 'json') {
-                push @l, '$_pci_r->{format} = (-t STDOUT) ? "json-pretty" : "json";', "\n";
-            } elsif ($co eq 'naked_res') {
-                push @l, '$_pci_r->{naked_res} = 1;', "\n";
-            } elsif ($co eq 'no_naked_res') {
-                push @l, '$_pci_r->{naked_res} = 0;', "\n";
+        for my $stage (1, 2) {
+            if ($stage == 1) {
+                push @l, 'my $go_spec1 = {', "\n";
             } else {
-                die "BUG: Unrecognized common_opt '$co'";
+                push @l, 'my $go_spec2 = {', "\n";
+                push @l, "  options => {\n";
             }
 
-            push @l, "      },\n"; # end of handler
-            push @l, "    },\n"; # end of option
-        }
-        push @l, "  },\n"; # end of options
-
-        # subcommand options
-        push @l, "  subcommands => {\n";
-        for my $sc_name (sort keys %{ $cd->{metas} }) {
-            my $meta = $cd->{metas}{$sc_name};
-            push @l, "    '$sc_name' => {\n";
-            push @l, "      options => {\n";
-            my $ggl_res = $cd->{ggl_res}{$sc_name};
+            # common options
+            my $ggl_res = $cd->{ggl_res}{$sc_names[0]};
             my $specmetas = $ggl_res->[3]{'func.specmeta'};
             for my $o (sort keys %$specmetas) {
                 my $specmeta = $specmetas->{$o};
-                my $argname = $specmeta->{arg}; # XXX can't handle submetadata yet
-                next unless defined $argname;
-                my $arg_spec = $meta->{args}{$argname};
-                push @l, "        '$o' => {\n";
-                push @l, "          handler => sub { ";
-                if ($specmeta->{is_alias} && $specmeta->{is_code}) {
-                    my $alias_spec = $arg_spec->{cmdline_aliases}{$specmeta->{alias}};
-                    if ($specmeta->{is_code}) {
-                        push @l, 'my $code = ', dmp($alias_spec->{code}), '; ';
-                        push @l, '$code->(\%_pci_args);';
-                    } else {
-                        push @l, '$_pci_args{\'', $specmeta->{arg}, '\'} = $_[1];';
-                    }
+                my $co = $specmeta->{common_opt};
+                next unless $co;
+                if ($stage == 1) {
+                    push @l, "  '$o' => sub { ", _gen_common_opt_handler($cd, $co), " },\n";
                 } else {
-                    if (($specmeta->{parsed}{type} // '') =~ /\@/) {
-                        push @l, 'if ($mentioned_args{\'', $specmeta->{arg}, '\'}++) { push @{ $_pci_args{\'', $specmeta->{arg}, '\'} }, $_[1] } else { $_pci_args{\'', $specmeta->{arg}, '\'} = [$_[1]] }';
-                    } elsif ($specmeta->{is_json}) {
-                        push @l, '$_pci_args{\'', $specmeta->{arg}, '\'} = _pci_json()->decode($_[1]);';
-                        _add_module($cd, "JSON::Tiny::Subclassable");
+                    push @l, "    '$o' => {\n";
+                    if ($co eq 'cmd') {
+                        push @l, "      handler => sub { ", _gen_common_opt_handler($cd, $co), " },\n";
                     } else {
-                        push @l, '$_pci_args{\'', $specmeta->{arg}, '\'} = $_[1];';
+                        push @l, "      handler => sub {},\n";
                     }
+                    push @l, "    },\n";
                 }
-                push @l, " },\n"; # end of handler
-                push @l, "        },\n"; # end of option
             }
-            push @l, "      },\n"; # end of options
-            push @l, "    },\n"; # end of subcommand
-        }
-        push @l, "  },\n"; # end of subcommands
-        push @l, "  default_subcommand => ".dmp($cd->{gen_args}{default_subcommand}).",\n";
+            if ($stage == 1) {
+                push @l, "};\n"; # end of %go_spec1
+            } else {
+                push @l, "  },\n"; # end of options
+            }
 
-        push @l, ");\n";
+            if ($stage == 2) {
+                # subcommand options
+                push @l, "  subcommands => {\n";
+                for my $sc_name (sort keys %{ $cd->{metas} }) {
+                    my $meta = $cd->{metas}{$sc_name};
+                    push @l, "    '$sc_name' => {\n";
+                    push @l, "      options => {\n";
+                    my $ggl_res = $cd->{ggl_res}{$sc_name};
+                    my $specmetas = $ggl_res->[3]{'func.specmeta'};
+                    for my $o (sort keys %$specmetas) {
+                        my $specmeta = $specmetas->{$o};
+                        my $argname = $specmeta->{arg}; # XXX can't handle submetadata yet
+                        next unless defined $argname;
+                        my $arg_spec = $meta->{args}{$argname};
+                        push @l, "        '$o' => {\n";
+                        push @l, "          handler => sub { ";
+                        if ($specmeta->{is_alias} && $specmeta->{is_code}) {
+                            my $alias_spec = $arg_spec->{cmdline_aliases}{$specmeta->{alias}};
+                            if ($specmeta->{is_code}) {
+                                push @l, 'my $code = ', dmp($alias_spec->{code}), '; ';
+                                push @l, '$code->(\%_pci_args);';
+                            } else {
+                                push @l, '$_pci_args{\'', $specmeta->{arg}, '\'} = $_[1];';
+                            }
+                        } else {
+                            if (($specmeta->{parsed}{type} // '') =~ /\@/) {
+                                push @l, 'if ($mentioned_args{\'', $specmeta->{arg}, '\'}++) { push @{ $_pci_args{\'', $specmeta->{arg}, '\'} }, $_[1] } else { $_pci_args{\'', $specmeta->{arg}, '\'} = [$_[1]] }';
+                            } elsif ($specmeta->{is_json}) {
+                                push @l, '$_pci_args{\'', $specmeta->{arg}, '\'} = _pci_json()->decode($_[1]);';
+                                _add_module($cd, "JSON::Tiny::Subclassable");
+                            } else {
+                                push @l, '$_pci_args{\'', $specmeta->{arg}, '\'} = $_[1];';
+                            }
+                        }
+                        push @l, " },\n"; # end of handler
+                        push @l, "        },\n"; # end of option
+                    }
+                    push @l, "      },\n"; # end of options
+                    push @l, "    },\n"; # end of subcommand
+                }
+                push @l, "  },\n"; # end of subcommands
+                push @l, "  default_subcommand => ".dmp($cd->{gen_args}{default_subcommand}).",\n";
+
+                push @l, "};\n"; # end of %go_spec2
+            } # subcommand options
+        } # stage
+
+        push @l, "{\n";
+        push @l, '  local @ARGV = @ARGV;', "\n";
+        push @l, '  my $old_conf = Getopt::Long::EvenLess::Configure("pass_through");', "\n";
+        push @l, '  Getopt::Long::EvenLess::GetOptions(%$go_spec1);', "\n";
+        push @l, '  Getopt::Long::EvenLess::Configure($old_conf);', "\n";
+        push @l, '  { my $first_non_opt_arg; for (@ARGV) { next if /^-/; $first_non_opt_arg = $_; last } if (!length $_pci_r->{subcommand_name} && defined $first_non_opt_arg) { $_pci_r->{subcommand_name} = $first_non_opt_arg } }', "\n";
+        push @l, '  if (!length $_pci_r->{subcommand_name}) { $_pci_r->{subcommand_name} = '.dmp($cd->{gen_args}{default_subcommand}).' } ' if defined $cd->{gen_args}{default_subcommand};
+        push @l, "}\n";
+        push @l, _gen_read_env($cd);
+        push @l, _gen_read_config($cd);
+        push @l, 'my $res = Getopt::Long::Subcommand::GetOptions(%$go_spec2);', "\n";
         push @l, '_pci_debug("args after GetOptions: ", \%_pci_args);', "\n" if $cd->{gen_args}{with_debug};
         push @l, '_pci_err([500, "GetOptions failed"]) unless $res->{success};', "\n";
-        push @l, 'if (!length $_pci_r->{subcommand_name} && @{ $res->{subcommand} // [] }) { $_pci_r->{subcommand_name} = $res->{subcommand}[0] }', "\n";
-        push @l, 'if (!length $_pci_r->{subcommand_name}) { $_pci_r->{subcommand_name} = '.dmp($cd->{gen_args}{default_subcommand}).' } ' if defined $cd->{gen_args}{default_subcommand};
         push @l, 'if (!length $_pci_r->{subcommand_name}) { print $help_msg; exit 0 }', "\n";
 
     } else {
@@ -468,79 +597,57 @@ sub _gen_parse_opts {
         _add_module($cd, "Getopt::Long::EvenLess");
         push @l, "require Getopt::Long::EvenLess;\n";
         my $meta = $cd->{metas}{''};
-        push @l, 'my $go_spec = {', "\n";
-        for my $go_spec (sort keys %{ $cd->{ggl_res}{''}[2] }) {
-            my $specmeta = $cd->{ggl_res}{''}[3]{'func.specmeta'}{$go_spec};
-            push @l, "    '$go_spec' => sub {\n";
-            my $co = $specmeta->{common_opt};
-            if ($co) {
-                if ($co eq 'help') {
-                    require Perinci::CmdLine::Help;
-                    my $res = Perinci::CmdLine::Help::gen_help(
-                        meta => $meta,
-                        common_opts => $cd->{copts},
-                        program_name => $cd->{script_name},
-                    );
-                    return [500, "Can't generate help: $res->[0] - $res->[1]"]
-                        unless $res->[0] == 200;
-                    push @l, '        print ', dmp($res->[2]), '; exit 0;', "\n";
-                } elsif ($co eq 'version') {
-                    # begin A, duplicated by B
-                    no strict 'refs';
-                    my $mod = $cd->{sc_mods}{''};
-                    push @l, "        no warnings 'once';\n";
-                    push @l, "        require $mod;\n" if $mod;
-                    push @l, '        print "', $cd->{script_name} , ' version ", ',
-                        (defined($cd->{gen_args}{script_version}) ? "\"$cd->{gen_args}{script_version}\"" :
-                         "(\$$mod\::VERSION // '?')"),
-                         ", (\$$mod\::DATE ? \" (\$$mod\::DATE)\" : '')",
-                         ', "\\n";', "\n";
-                    push @l, '        print "  Generated by ', __PACKAGE__ , ' version ',
-                        (${__PACKAGE__."::VERSION"} // 'dev'),
-                        (${__PACKAGE__."::DATE"} ? " (".${__PACKAGE__."::DATE"}.")" : ""),
-                        '\n";', "\n";
-                    push @l, '        exit 0;', "\n";
-                    # end A
-                } elsif ($co eq 'format') {
-                    push @l, '        $_pci_r->{format} = $_[1];', "\n";
-                } elsif ($co eq 'json') {
-                    push @l, '        $_pci_r->{format} = (-t STDOUT) ? "json-pretty" : "json";', "\n";
-                } elsif ($co eq 'naked_res') {
-                    push @l, '        $_pci_r->{naked_res} = 1;', "\n";
-                } elsif ($co eq 'no_naked_res') {
-                    push @l, '        $_pci_r->{naked_res} = 0;', "\n";
-                } else {
-                    die "BUG: Unrecognized common_opt '$co'";
-                }
-            } else {
-                my $arg_spec = $meta->{args}{$specmeta->{arg}};
-                push @l, '        ';
-                if ($specmeta->{is_alias} && $specmeta->{is_code}) {
-                    my $alias_spec = $arg_spec->{cmdline_aliases}{$specmeta->{alias}};
-                    if ($specmeta->{is_code}) {
-                        push @l, 'my $code = ', dmp($alias_spec->{code}), '; ';
-                        push @l, '$code->(\%_pci_args);';
+        # stage 1 is catching common options only (--help, etc)
+        for my $stage (1, 2) {
+            push @l, "my \$go_spec$stage = {\n";
+            for my $go_spec (sort keys %{ $cd->{ggl_res}{''}[2] }) {
+                my $specmeta = $cd->{ggl_res}{''}[3]{'func.specmeta'}{$go_spec};
+                my $co = $specmeta->{common_opt};
+                next if $stage == 1 && !$co;
+                push @l, "    '$go_spec' => sub { "; # begin option handler
+                if ($co) {
+                    if ($stage == 1) {
+                        push @l, _gen_common_opt_handler($cd, $co);
                     } else {
-                        push @l, '$_pci_args{\'', $specmeta->{arg}, '\'} = $_[1];';
+                        # empty, we've done handling common options in stage 1
                     }
                 } else {
-                    if (($specmeta->{parsed}{type} // '') =~ /\@/) {
-                        push @l, 'if ($mentioned_args{\'', $specmeta->{arg}, '\'}++) { push @{ $_pci_args{\'', $specmeta->{arg}, '\'} }, $_[1] } else { $_pci_args{\'', $specmeta->{arg}, '\'} = [$_[1]] }';
-                    } elsif ($specmeta->{is_json}) {
-                        push @l, '$_pci_args{\'', $specmeta->{arg}, '\'} = _pci_json()->decode($_[1]);';
-                        _add_module($cd, "JSON::Tiny::Subclassable");
+                    my $arg_spec = $meta->{args}{$specmeta->{arg}};
+                    push @l, '        ';
+                    if ($stage == 1) {
+                        # in stage 1, we do not yet deal with argument options
+                    } elsif ($specmeta->{is_alias} && $specmeta->{is_code}) {
+                        my $alias_spec = $arg_spec->{cmdline_aliases}{$specmeta->{alias}};
+                        if ($specmeta->{is_code}) {
+                            push @l, 'my $code = ', dmp($alias_spec->{code}), '; ';
+                            push @l, '$code->(\%_pci_args);';
+                        } else {
+                            push @l, '$_pci_args{\'', $specmeta->{arg}, '\'} = $_[1];';
+                        }
                     } else {
-                        push @l, '$_pci_args{\'', $specmeta->{arg}, '\'} = $_[1];';
+                        if (($specmeta->{parsed}{type} // '') =~ /\@/) {
+                            push @l, 'if ($mentioned_args{\'', $specmeta->{arg}, '\'}++) { push @{ $_pci_args{\'', $specmeta->{arg}, '\'} }, $_[1] } else { $_pci_args{\'', $specmeta->{arg}, '\'} = [$_[1]] }';
+                        } elsif ($specmeta->{is_json}) {
+                            push @l, '$_pci_args{\'', $specmeta->{arg}, '\'} = _pci_json()->decode($_[1]);';
+                            _add_module($cd, "JSON::Tiny::Subclassable");
+                        } else {
+                            push @l, '$_pci_args{\'', $specmeta->{arg}, '\'} = $_[1];';
+                        }
                     }
+                    push @l, "\n";
                 }
-                push @l, "\n";
-            }
-            push @l, "    },\n";
-        }
-        push @l, "};\n";
-        push @l, 'my $res = Getopt::Long::EvenLess::GetOptions(%$go_spec);', "\n";
-        push @l, '_pci_debug("args after GetOptions: ", \%_pci_args);', "\n" if $cd->{gen_args}{with_debug};
+                push @l, " },\n"; # end option handler
+            } # options
+            push @l, "};\n";
+        } # stage
+        push @l, 'my $old_conf = Getopt::Long::EvenLess::Configure("pass_through");', "\n";
+        push @l, 'Getopt::Long::EvenLess::GetOptions(%$go_spec1);', "\n";
+        push @l, 'Getopt::Long::EvenLess::Configure($old_conf);', "\n";
+        push @l, _gen_read_env($cd);
+        push @l, _gen_read_config($cd);
+        push @l, 'my $res = Getopt::Long::EvenLess::GetOptions(%$go_spec2);', "\n";
         push @l, '_pci_err([500, "GetOptions failed"]) unless $res;', "\n";
+        push @l, '_pci_debug("args after GetOptions (stage 2): ", \%_pci_args);', "\n" if $cd->{gen_args}{with_debug};
 
     }
 
@@ -556,12 +663,11 @@ my %pericmd_attrs = (
             schema => 'any*',
         },
     )} qw/actions common_opts completion
+          default_format
           description exit formats
           riap_client riap_version riap_client_args
           log
           tags
-          read_env env_name
-          read_config config_dirs config_filename
           get_subcommand_from_arg
          /),
 
@@ -647,7 +753,7 @@ _
         ],
         'req_one&' => [
             [qw/url meta/],
-            [qw/meta subcommands/],
+            [qw/url subcommands/],
         ],
         'choose_all&' => [
             [qw/meta sub_name/],
@@ -690,15 +796,42 @@ _
             schema  => 'str*',
         },
         validate_args => {
-            summary => 'Whether to validate arguments using schemas',
+            summary => 'Whether the CLI script should validate arguments using schemas',
             schema  => 'bool',
             default => 1,
         },
         #validate_result => {
-        #    summary => 'Whether to validate result using schemas',
+        #    summary => 'Whether the CLI script should validate result using schemas',
         #    schema  => 'bool',
         #    default => 1,
         #},
+        read_config => {
+            summary => 'Whether the CLI script should read configuration files',
+            schema => 'bool*',
+            default => 1,
+        },
+        config_filename => {
+            summary => 'Configuration file name(s)',
+            schema => ['any*', of=>[
+                'str*',
+                'hash*',
+                ['array*', of=>['any*', of=>['str*','hash*']]],
+            ]],
+        },
+        config_dirs => {
+            'x.name.is_plural' => 1,
+            'x.name.singular' => 'config_dir',
+            summary => 'Where to search for configuration files',
+            schema => ['array*', of=>'str*'],
+        },
+        read_env => {
+            summary => 'Whether CLI script should read environment variable that sets default options',
+            schema => 'bool*',
+        },
+        env_name => {
+            summary => 'Name of environment variable name that sets default options',
+            schema => 'str*',
+        },
 
         with_debug => {
             summary => 'Generate script with debugging outputs',
@@ -741,7 +874,7 @@ _
 
         pack_deps => {
             summary => 'Whether to pack dependencies into the script',
-            schema => ['bool*', is=>1],
+            schema => ['bool*'],
             default => 1,
             description => <<'_',
 
@@ -756,7 +889,7 @@ _
         },
         pod => {
             summary => 'Whether to generate POD for the script',
-            schema => ['bool*', is=>1],
+            schema => ['bool*'],
             default => 1,
         },
 
@@ -782,6 +915,8 @@ sub gen_inline_pericmd_script {
     $args{validate_args} //= 1;
     #$args{validate_result} //= 1;
     $args{pack_deps} //= 1;
+    $args{read_config} //= 1;
+    $args{read_env} //= 1;
 
     my $cd = {
         gen_args => \%args,
@@ -865,6 +1000,14 @@ sub gen_inline_pericmd_script {
         $cd->{func_names} = \%func_names;
     } # GET_META
 
+    $args{config_filename} //= "$cd->{script_name}.conf";
+    $args{env_name} //= do {
+        my $env = uc "$cd->{script_name}_OPT";
+        $env =~ s/[^A-Z0-9]+/_/g;
+        $env = "_$env" if $env =~ /\A\d/;
+        $env;
+    };
+
     for (
         # required by Perinci::Result::Format::Lite. this will be removed if we
         # don't need formatting.
@@ -917,6 +1060,16 @@ sub gen_inline_pericmd_script {
                 $copts{subcommands} = $Perinci::CmdLine::Base::copts{subcommands};
                 $copts{cmd}         = $Perinci::CmdLine::Base::copts{cmd};
             }
+            if ($args{read_config}) {
+                for (qw/config_path no_config config_profile/) {
+                    $copts{$_} = $Perinci::CmdLine::Base::copts{$_};
+                }
+            }
+            if ($args{read_env}) {
+                for (qw/no_env/) {
+                    $copts{$_} = $Perinci::CmdLine::Base::copts{$_};
+                }
+            }
             $cd->{copts} = \%copts;
         }
 
@@ -1023,15 +1176,19 @@ _
             _gen_pci_check_args($cd);
         }
 
-        $cd->{vars}{'$_pci_r'} = {naked_res=>0, subcommand_name=>''};
+        $cd->{vars}{'$_pci_r'} = {
+            naked_res => 0,
+            subcommand_name => '',
+            read_config => $args{read_config},
+            read_env => $args{read_env},
+        };
 
-        # gen code to parse cmdline options
         $cd->{vars}{'%_pci_args'} = undef;
-        push @l, "### parse cmdline options\n\n";
-        push @l, "{\n", _gen_parse_opts($cd), "}\n\n";
+        push @l, "### get arguments (from config file, env, command-line args\n\n";
+        push @l, "{\n", _gen_get_args($cd), "}\n\n";
 
         # gen code to check arguments
-        push @l, "### parse cmdline options\n\n";
+        push @l, "### check arguments\n\n";
         push @l, "{\n";
         push @l, 'require Local::_pci_check_args; ' if $cd->{gen_args}{pack_deps};
         push @l, 'my $res = _pci_check_args(\\%_pci_args);', "\n";
@@ -1156,11 +1313,11 @@ _
                 default_subcommand => $args{default_subcommand},
                 per_arg_json       => 1,
                 per_arg_yaml       => 0,
-                read_env           => 0, # TODO
-                #env_name           => $args{env_name}, TODO
-                read_config        => 0, # TODO
-                #config_filename    => $args{config_filenames},
-                #config_dirs        => $args{config_dirs},
+                read_env           => $args{read_env},
+                env_name           => $args{env_name},
+                read_config        => $args{read_config},
+                config_filename    => $args{config_filenames},
+                config_dirs        => $args{config_dirs},
                 completer_script    => "_$cd->{script_name}",
             );
             return err($res, 500, "Can't generate POD") unless $res->[0] == 200;
@@ -1183,11 +1340,10 @@ _
                 JSON::MaybeXS::encode_json(\%tmp);
             }, "\n\n",
 
-            "# PERICMD_INLINE_SCRIPT_METAS: ", do {
-                my $metas = Data::Clean::JSON->get_cleanser->clone_and_clean(
-                    $cd->{metas});
-                JSON::MaybeXS::encode_json($metas);
-            }, "\n\n",
+            'my %_pci_metas = %{ ', do {
+                local $Data::Dmp::OPT_DEPARSE=0;
+                dmp($cd->{metas});
+            }, " };\n\n",
 
             "# This script is generated by ", __PACKAGE__,
             " version ", (${__PACKAGE__."::VERSION"} // 'dev'), " on ",
@@ -1291,7 +1447,7 @@ Perinci::CmdLine::Inline - Generate inline Perinci::CmdLine CLI script
 
 =head1 VERSION
 
-This document describes version 0.44 of Perinci::CmdLine::Inline (from Perl distribution Perinci-CmdLine-Inline), released on 2017-01-11.
+This document describes version 0.45 of Perinci::CmdLine::Inline (from Perl distribution Perinci-CmdLine-Inline), released on 2017-01-12.
 
 =head1 SYNOPSIS
 
@@ -1412,11 +1568,15 @@ Currently does nothing, provided only for compatibility with Perinci::CmdLine::B
 
 Currently does nothing, provided only for compatibility with Perinci::CmdLine::Base.
 
-=item * B<config_dirs> => I<any>
+=item * B<config_dirs> => I<array[str]>
 
-Currently does nothing, provided only for compatibility with Perinci::CmdLine::Base.
+Where to search for configuration files.
 
-=item * B<config_filename> => I<any>
+=item * B<config_filename> => I<str|hash|array[str|hash]>
+
+Configuration file name(s).
+
+=item * B<default_format> => I<any>
 
 Currently does nothing, provided only for compatibility with Perinci::CmdLine::Base.
 
@@ -1426,9 +1586,9 @@ Currently does nothing, provided only for compatibility with Perinci::CmdLine::B
 
 Currently does nothing, provided only for compatibility with Perinci::CmdLine::Base.
 
-=item * B<env_name> => I<any>
+=item * B<env_name> => I<str>
 
-Currently does nothing, provided only for compatibility with Perinci::CmdLine::Base.
+Name of environment variable name that sets default options.
 
 =item * B<exit> => I<any>
 
@@ -1487,13 +1647,13 @@ Whether to pass Perinci::CmdLine::Inline object.
 
 Whether to generate POD for the script.
 
-=item * B<read_config> => I<any>
+=item * B<read_config> => I<bool> (default: 1)
 
-Currently does nothing, provided only for compatibility with Perinci::CmdLine::Base.
+Whether the CLI script should read configuration files.
 
-=item * B<read_env> => I<any>
+=item * B<read_env> => I<bool>
 
-Currently does nothing, provided only for compatibility with Perinci::CmdLine::Base.
+Whether CLI script should read environment variable that sets default options.
 
 =item * B<riap_client> => I<any>
 
@@ -1545,7 +1705,7 @@ Whether to set utf8 flag on output.
 
 =item * B<validate_args> => I<bool> (default: 1)
 
-Whether to validate arguments using schemas.
+Whether the CLI script should validate arguments using schemas.
 
 =item * B<with_debug> => I<bool>
 
